"""Game state."""

from common import state
import enum
import pygame
from pygame.locals import *
from common import color  # needs to come after pygame.locals.*
from . import room


class Game(state.GameState):

    def __init__(self, screen):
        self.view = room.View.DEFAULT
        self._images = room.Images(screen)
        self._keypad_test = room.KeyPadTest(
            self._images.keypad, self._images.door)
        self._wall_color = color.GREY
        super().__init__(screen)

    def _draw_default(self):
        pygame.draw.rect(self.screen, color.BLACK, room.BACK_WALL, 5)
        for corner in ('topleft', 'bottomleft', 'bottomright', 'topright'):
            pygame.draw.line(
                self.screen, color.BLACK, getattr(room.RECT, corner),
                getattr(room.BACK_WALL, corner), 5)
        if not self._images.light_switch.on:
            self._images.mini_zodiac.draw()
        self._images.mini_window.draw()
        self._images.mini_math.draw()
        self._images.mini_chest.draw()
        self._images.mini_light_switch.draw()

    def _draw_back_wall(self):
        self._images.math.draw()

    def _draw_front_wall(self):
        self._images.door.draw()

    def _draw_front_keypad(self):
        self.screen.fill(color.DARK_GREY_2)
        self._images.keypad.draw()

    def _draw_floor(self):
        self._images.chest.draw()

    def _draw_ceiling(self):
        if not self._images.light_switch.on:
            self._images.zodiac.draw()

    def _draw_left_wall(self):
        self._images.window.draw()

    def _draw_left_window(self):
        self._images.maxi_window.draw()

    def _draw_right_wall(self):
        self._images.light_switch.draw()

    def draw(self):
        """Draw the current view.

        First fills in the wall color, then calls _draw_<lowercase view name>.
        """
        self.screen.fill(self._wall_color)
        if self._images.keypad.opened and self.view != room.View.FRONT_KEYPAD:
            # Navigating away from the keypad after opening resets it.
            self._keypad_test.stop()
        view = self.view.name.lower()
        getattr(self, f'_draw_{view}')()
        pygame.display.update()

    def _handle_default_click(self, pos):
        if room.at_edge(pos):
            # In the default view, the edges of the screen border the (hidden)
            # front wall.
            self.view = room.View.FRONT_WALL
        elif self._images.mini_chest.collidepoint(pos):
            # Clicking in the intersection of the chest on the floor and the
            # back wall should take us to the floor, not the wall.
            self.view = room.View.FLOOR
        elif room.BACK_WALL.collidepoint(pos):
            # The back wall is a rectangle, so use built-in collision detection.
            self.view = room.View.BACK_WALL
        else:
            # Since we've already checked the front and back walls, the quadrant
            # corresponds to one of the other four walls.
            self.view = room.View(room.quadrant(pos).value)
        return True

    def _handle_generic_click(self, pos):
        if not room.at_edge(pos):
            return False
        quadrant = room.quadrant(pos)
        self.view = room.View(
            room.ROTATIONS[self.view].get(quadrant, quadrant.value))
        return True

    def _handle_left_wall_click(self, pos):
        if not self._images.window.collidepoint(pos):
            return False
        self.view = room.View.LEFT_WINDOW
        return True

    def _handle_front_wall_click(self, pos):
        if not self._images.door.collidepoint(pos):
            return False
        if self._images.door.revealed:
            self.view = room.View.FRONT_KEYPAD
        else:
            self._images.door.revealed = True
        return True

    def _toggle_light_switch(self):
        self._images.light_switch.on = not self._images.light_switch.on
        self._images.mini_light_switch.on = self._images.light_switch.on
        if self._images.light_switch.on:
            self._wall_color = color.GREY
        else:
            self._wall_color = color.DARK_GREY_2
        self._images.door.light_switch_on = self._images.light_switch.on

    def _handle_right_wall_click(self, pos):
        if not self._images.light_switch.collidepoint(pos):
            return False
        self._toggle_light_switch()
        return True

    def handle_click(self, event):
        """Handle left mouse clicks.

        If the event is a left click, tries to call
        _handle_{lowercase view name}. If it does not exist or returns False,
        falls back to _handle_generic_click.

        Args:
          event: An event.
        """
        if event.type != MOUSEBUTTONDOWN or event.button != 1:
            return False
        view = self.view.name.lower()
        handle_view_click = getattr(self, f'_handle_{view}_click', None)
        if handle_view_click:
            consumed = handle_view_click(event.pos)
        else:
            consumed = False
        if not consumed:
            consumed = self._handle_generic_click(event.pos)
        if consumed:
            self.draw()
        return consumed

    def handle_reset(self, event):
        if not state.keypressed(event, K_ESCAPE):
            return False
        if self.view is not room.View.DEFAULT:
            self.view = room.View.DEFAULT
            self.draw()
        return True

    def handle_chest_combo(self, event):
        if self.view is not room.View.FLOOR:
            return False
        response = self._images.chest.send(event)
        if response is None:
            return False
        if response:
            if self._images.chest.opened:
                self.draw()
            else:
                # Optimization: only redraw the chest area.
                self._images.chest.draw()
                pygame.display.update()
            self._images.mini_chest.text = self._images.chest.text
        return True

    def handle_keypad_input(self, event):
        if self.view is not room.View.FRONT_KEYPAD:
            return False
        response = self._images.keypad.send(event)
        if response is None:
            return False
        if response:
            self._images.keypad.draw()
            pygame.display.update()
            self._images.door.text = self._images.keypad.text
        return True

    def handle_keypad_test(self, event):
        if not self._images.keypad.opened:
            return False
        consumed = self._keypad_test.send(event)
        if consumed:
            self._images.keypad.draw()
            pygame.display.update()
        if self._keypad_test.completed:
            self.active = False
        return consumed

    def ending(self):
        assert not self.active
        if self._keypad_test.completed:
            return Ending(self.screen, self._wall_color, self._images.keypad)
        return None


class _EndingFrame(enum.IntEnum):
    """Frames of the ending sequence.

    Each frame is displayed for
    (next_frame - current_frame) * Ending._WAIT_INTERVAL_MS milliseconds.
    """
    KEYPAD = 0
    KEYPAD_BLUE = 1
    DOOR = 2
    CONGRATS = 4
    DETAIL = 5
    WARNING = 7
    FIN = 9


class Ending(state.GameState):

    _TICK = pygame.USEREVENT
    _WAIT_INTERVAL_MS = 1250

    _OPEN_DOOR = [room.DOOR_RECT.topleft,
                  (room.DOOR_RECT.left + 20, room.DOOR_RECT.top + 20),
                  (room.DOOR_RECT.left + 20, room.DOOR_RECT.bottom - 10),
                  room.DOOR_RECT.bottomleft]

    def __init__(self, screen, wall_color, keypad):
        self._wall_color = wall_color
        self._keypad = keypad
        self._frame = 0
        self._font = pygame.font.SysFont('couriernew', 100, bold=True)
        super().__init__(screen)
        pygame.time.set_timer(self._TICK, self._WAIT_INTERVAL_MS)

    def _render_text(self, text, raw_pos):
        surface = self._font.render(text, 0, color.BLACK)
        size = surface.get_size()
        pos = tuple(raw_pos[i] - size[i] / 2 for i in range(2))
        self.screen.blit(surface, pos)

    def draw(self):
        if self._frame in (_EndingFrame.KEYPAD, _EndingFrame.KEYPAD_BLUE):
            self.screen.fill(color.DARK_GREY_2)
            self._keypad.draw()
        elif self._frame < _EndingFrame.CONGRATS:
            self.screen.fill(self._wall_color)
            pygame.draw.rect(self.screen, color.BLUE, room.DOOR_RECT)
            pygame.draw.polygon(self.screen, color.DARK_GREY_2, self._OPEN_DOOR)
            pygame.draw.rect(self.screen, color.BLACK, room.DOOR_RECT, 5)
        else:
            self.screen.fill(color.BLUE)
            self._render_text('Congratulations',
                              (room.RECT.w / 2, room.RECT.h / 4))
            if self._frame >= _EndingFrame.DETAIL:
                self._render_text('you escaped',
                                  (room.RECT.w / 2, room.RECT.h / 2))
            if self._frame >= _EndingFrame.WARNING:
                self._render_text('for now...',
                                  (room.RECT.w / 2, 3 * room.RECT.h / 4))
        pygame.display.update()

    def handle_tick(self, event):
        if event.type != self._TICK:
            return False
        self._frame += 1
        if self._frame == _EndingFrame.FIN:
            self.active = False
            return True
        elif self._frame == _EndingFrame.KEYPAD_BLUE:
            self._keypad.text_color = color.BLUE
        self.draw()
        return True

    def cleanup(self):
        pygame.time.set_timer(self._TICK, 0)



"""Game room."""

import abc
from common import img
from common import state as common_state
import enum
import pygame
from pygame.locals import *
from common import color  # needs to come after pygame.locals.*
import random
import string
from typing import Dict, List, NamedTuple, Tuple


RECT = common_state.RECT
BACK_WALL = pygame.Rect(RECT.w / 4, RECT.h / 4, RECT.w / 2, RECT.h / 2)
DOOR_RECT = pygame.Rect(2 * RECT.w / 5, RECT.h / 4, RECT.w / 5, 3 * RECT.h / 4)


class View(enum.Enum):
    LEFT_WALL = 0
    CEILING = 1
    RIGHT_WALL = 2
    FLOOR = 3
    FRONT_WALL = 4
    BACK_WALL = 5
    DEFAULT = 6
    LEFT_WINDOW = 7
    FRONT_KEYPAD = 8


class Quadrant(enum.Enum):
    LEFT = 0
    TOP = 1
    RIGHT = 2
    BOTTOM = 3


ROTATIONS = {View.LEFT_WALL: {Quadrant.LEFT: View.FRONT_WALL.value,
                              Quadrant.RIGHT: View.BACK_WALL.value},
             View.CEILING: {Quadrant.TOP: View.FRONT_WALL.value,
                            Quadrant.BOTTOM: View.BACK_WALL.value},
             View.RIGHT_WALL: {Quadrant.LEFT: View.BACK_WALL.value,
                               Quadrant.RIGHT: View.FRONT_WALL.value},
             View.FLOOR: {Quadrant.TOP: View.BACK_WALL.value,
                          Quadrant.BOTTOM: View.FRONT_WALL.value}}
ROTATIONS[View.FRONT_WALL] = {quad: (2 - quad.value) % 4 for quad in Quadrant}
ROTATIONS[View.BACK_WALL] = ROTATIONS[View.DEFAULT] = {}
ROTATIONS[View.LEFT_WINDOW] = {quad: View.LEFT_WALL for quad in Quadrant}
ROTATIONS[View.FRONT_KEYPAD] = {quad: View.FRONT_WALL for quad in Quadrant}


def at_edge(pos):
    x, y = pos
    return x <= 10 or y <= 10 or RECT.w - x <= 10 or RECT.h - y <= 10


def quadrant(pos):
    x, y = pos
    if x == 0:
        # To avoid divide-by-zero, special-case the leftmost edge.
        return Quadrant.LEFT
    # Compute the (absolute values of) the slopes of the given position from the
    # top and bottom left corners of the screen.
    down_slope = y / x
    up_slope = (RECT.h - y) / x
    # Compute the slope of the screen.
    wall_slope = RECT.h / RECT.w
    # Determine the position relative to the screen diagonals.
    above_down_diagonal = down_slope < wall_slope
    above_up_diagonal = up_slope > wall_slope
    # The diagonals divide the screen into four quadrants. Use the relative
    # positions to determine which quadrant we're in.
    if not above_down_diagonal and above_up_diagonal:
        return Quadrant.LEFT
    if above_down_diagonal and above_up_diagonal:
        return Quadrant.TOP
    if above_down_diagonal and not above_up_diagonal:
        return Quadrant.RIGHT
    assert not above_down_diagonal and not above_up_diagonal
    return Quadrant.BOTTOM


def _font(size):
    return pygame.font.SysFont('couriernew', size, bold=True)


class _TextMixin(abc.ABC):

    _MAX_TEXT_LENGTH: int
    text: str

    @abc.abstractmethod
    def accept_event(self, event):
        raise NotImplementedError()

    @abc.abstractmethod
    def to_char(self, event):
        raise NotImplementedError()

    def send(self, event):
        """Sends an event that may be consumed as text input.

        Args:
          event: An event.

        Returns:
          None if the event should not be consumed.
          False if the event is consumed but a redraw is not required.
          True if the event is consumed and a redraw is required.
        """
        if not self.accept_event(event):
            return None
        if event.key == K_BACKSPACE:
            # Backspace is consumed as deletion of the rightmost character;
            # whether it requires a redraw depends on whether there exists a
            # character to delete.
            redraw = bool(self.text)
            self.text = self.text[:-1]
            return redraw
        c = self.to_char(event)
        if not c:
            return None
        if len(self.text) >= self._MAX_TEXT_LENGTH:
            return False
        self.text += c
        return True


class _ChestBase(img.RectFactory):

    _CHARPOS: List[Tuple[int, int]]
    _FONT_SIZE: int

    def __init__(self, names, screen, position, shift):
        super().__init__(screen)
        self.text = ''
        self._font = _font(self._FONT_SIZE)
        self._images = [
            img.load(name, screen, position, shift) for name in names]
        self.RECT = self._images[self.opened].RECT

    @property
    def opened(self):
        return self.text == 'AYP'

    def draw(self):
        self._images[self.opened].draw()
        if not self.opened:
            for c, pos in zip(self.text, self._CHARPOS):
                self._screen.blit(self._font.render(c, 0, color.BLACK), pos)


class Chest(_TextMixin, _ChestBase):

    _CHARPOS = [(496, 278), (509, 278), (521, 278)]
    _FONT_SIZE = 15
    _MAX_TEXT_LENGTH = 3

    def __init__(self, screen):
        # pytype: disable=wrong-arg-count
        super().__init__(('chest', 'chest_opened'), screen,
                         (RECT.w / 2, 2 * RECT.h / 3), (-0.5, -1))
        # pytype: enable=wrong-arg-count

    def accept_event(self, event):
        return event.type == KEYDOWN and not self.opened

    def to_char(self, event):
        if not event.unicode or event.unicode not in string.printable:
            # Aside from backspace, only printable characters are consumed.
            return None
        return event.unicode.upper()


class MiniChest(_ChestBase):

    _CHARPOS = [(501, 430), (510, 430), (519, 430)]
    _FONT_SIZE = 10

    def __init__(self, screen):
        super().__init__(('mini_chest', 'mini_chest_opened'), screen,
                         (RECT.w / 2, RECT.h * 7 / 8), (-0.5, -1))


class _DigitsBase(img.Factory):

    _DIGITS: Dict[str, Tuple[int, int]]
    _FONT_SIZE: int

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)  # pytype: disable=wrong-arg-count
        self._font = _font(self._FONT_SIZE)
        self._digits = tuple(
            self._font.render(d, 0, color.BLACK) for d in self._DIGITS)

    def draw(self):
        super().draw()
        for i, pos in enumerate(self._DIGITS.values()):
            self._screen.blit(self._digits[i], pos)


class Door(_DigitsBase):

    _GAP = pygame.Rect(
        DOOR_RECT.right - 5, DOOR_RECT.top, 10, DOOR_RECT.h)

    _DIGITS = {
        '1': (570, 343),
        '2': (580, 343),
        '3': (589, 343),
        '4': (569, 355),
        '5': (579, 355),
        '6': (587, 355),
        '7': (569, 367),
        '8': (579, 367),
        '9': (587, 367),
        '0': (579, 378),
    }
    _FONT_SIZE = 10

    def __init__(self, screen):
        super().__init__(screen)
        self._door = img.load('door', screen, (RECT.w / 2, RECT.h), (-0.5, -1))
        self.revealed = False
        self.light_switch_on = True
        self.text = ''

    def draw(self):
        if self.revealed:
            self._door.draw()
            super().draw()
            self._screen.blit(
                self._font.render(self.text, 0, color.BLACK), (570, 325))
            return
        if self.light_switch_on:
            gap_color = color.DARK_GREY_1
        else:
            gap_color = color.DARK_GREY_2
        pygame.draw.rect(self._screen, gap_color, DOOR_RECT, 5)
        pygame.draw.rect(self._screen, gap_color, self._GAP)

    def collidepoint(self, pos):
        return (self._door if self.revealed else self._GAP).collidepoint(pos)


class LightSwitchBase(img.RectFactory):

    def __init__(self, names, screen, position, shift):
        assert len(names) == 2  # image names for off and on positions
        super().__init__(screen)
        self.on = True
        self._images = [
            img.load(name, screen, position, shift) for name in names]
        self.RECT = self._images[self.on].RECT

    def draw(self):
        self._images[self.on].draw()


class LightSwitch(LightSwitchBase):

    def __init__(self, screen):
        super().__init__(('light_switch_off', 'light_switch_on'), screen,
                         (RECT.w / 2, RECT.h / 2), (-0.5, -1))


class MiniLightSwitch(LightSwitchBase):

    def __init__(self, screen):
        super().__init__(('mini_light_switch_off', 'mini_light_switch_on'),
                         screen, (RECT.w * 7 / 8, RECT.h / 2), (-0.5, -1))


class _KeyPadText(NamedTuple):
    value: str
    pos: Tuple[int, int]
    font: pygame.font.Font
    color: Tuple[int, int, int]


class KeyPad(_DigitsBase, _TextMixin, img.PngFactory):

    _DIGITS = {
        '1': (417, 183),
        '2': (487, 183),
        '3': (557, 183),
        '4': (417, 270),
        '5': (487, 270),
        '6': (557, 270),
        '7': (417, 355),
        '8': (487, 355),
        '9': (557, 355),
        '0': (487, 442),
    }
    _FONT_SIZE = 85
    _MAX_TEXT_LENGTH = 4
    _TEXT_POS = (410, 50)

    def __init__(self, screen):
        super().__init__('keypad', screen, (RECT.w / 2, 0), (-0.5, 0))
        self.set_initial_text()

    def set_initial_text(self):
        # the text that renders on the keypad
        self._text = _KeyPadText(value='', pos=self._TEXT_POS, font=self._font,
                                 color=color.BLACK)
        # the current input for opening the keypad, differs from self.text when
        # the keypad is already open
        self._opening_input = ''

    def _resized_text(self, value):
        assert self.opened
        max_size = self._font.size(self._opening_input)
        font_size = self._FONT_SIZE
        font = self._font
        size = font.size(value)
        while any(dim > max_dim for dim, max_dim in zip(size, max_size)):
            font_size -= 10
            font = _font(font_size)
            size = font.size(value)
        pos = tuple(self._TEXT_POS[i] + (max_size[i] - size[i]) / 2
                    for i in range(2))
        return self._text._replace(value=value, pos=pos, font=font)

    @property
    def text(self):
        return self._text.value

    @text.setter
    def text(self, value):
        if self.opened:
            self._text = self._resized_text(value)
        else:
            self._text = self._text._replace(value=value)
            self._opening_input = value

    @property
    def text_color(self):
        return self._text.color

    @text_color.setter
    def text_color(self, color):
        self._text = self._text._replace(color=color)

    @property
    def opened(self):
        return self._opening_input == '9710'

    def accept_event(self, event):
        return event.type == KEYDOWN and not self.opened

    def to_char(self, event):
        return event.unicode if event.unicode.isdigit() else None

    def draw(self):
        super().draw()
        text = self._text.font.render(self.text, 0, self._text.color)
        self._screen.blit(text, self._text.pos)


class Images:

    def __init__(self, screen):
        self.chest = img.load(screen, factory=Chest)
        self.mini_chest = img.load(screen, factory=MiniChest)

        self.door = img.load(screen, factory=Door)

        self.keypad = img.load(screen, factory=KeyPad)

        self.light_switch = img.load(screen, factory=LightSwitch)
        self.mini_light_switch = img.load(screen, factory=MiniLightSwitch)

        self.math = img.load('math', screen)
        self.mini_math = img.load('mini_math', screen, BACK_WALL.topleft)

        self.window = img.load(
            'window', screen, (RECT.w / 4, RECT.h / 2), (0, -1))
        self.mini_window = img.load(
            'mini_window', screen, (RECT.w / 16, RECT.h / 2), (0, -1))
        self.maxi_window = img.load('maxi_window', screen, (0, RECT.h / 4))

        self.zodiac = img.load('zodiac', screen)
        self.mini_zodiac = img.load('mini_zodiac', screen)


class _QuestionState(enum.Enum):
    ACTIVE = ''
    OK = 'OK'
    ERR = 'ERR'


class _Question:

    def __init__(self, value):
        self.value = value
        self.state = _QuestionState.ACTIVE
        self._num_ticks = 0

    def solve(self, answer):
        if answer == self.value[-1]:
            self.state = _QuestionState.OK
        else:
            self.state = _QuestionState.ERR

    def tick(self):
        self._num_ticks += 1
        if self._num_ticks > 1:
            self.state = _QuestionState.ERR


class KeyPadTest:

    _TICK = pygame.USEREVENT
    _FREQ_MS = 500

    _OPERATOR_REPR = {int.__add__: '+', int.__mul__: '*', int.__sub__: '-',
                      int.__truediv__: '/'}
    _OPERATORS = tuple(_OPERATOR_REPR)

    _SPECIAL_QUESTION = ('1', '+', '1', '3')

    def __init__(self, keypad, door):
        self._keypad = keypad
        self._door = door
        self._init_state()

    def _init_state(self):
        self._active = False
        self._question = None
        self._special_question_position = random.randint(9, 14)
        self._num_questions_asked = 0

    def _generate_question(self):
        if self._num_questions_asked == self._special_question_position:
            return self._SPECIAL_QUESTION
        answer = random.randint(0, 9)
        operand1 = random.randint(0, 9)
        operator_idx = random.randint(0, 3)
        operator = self._OPERATORS[operator_idx]
        if not operand1 and operator_idx % 2:  # avoid 0 * x and x / 0
            return self._generate_question()
        elif answer == 2 and operand1 == 1 and not operator_idx:  # avoid 1 + 1
            return self._generate_question()
        # Use the inverse operator to compute the other operand.
        operand2 = self._OPERATORS[(operator_idx + 2) % 4](answer, operand1)
        if round(operand2, 1) != operand2:
            # Reject questions where the second operand has more than one digit
            # after the decimal point.
            return self._generate_question()
        if operator_idx > 1:  # for - and /, we need to flip the operands
            operand1, operand2 = operand2, operand1

        def to_str(n):
            # int / int produces a float - convert floats to ints when possible.
            s = str(int(n) if round(n, 0) == n else n)
            # Parenthesize negative numbers for readability.
            return f'({s})' if n < 0 else s

        return (to_str(operand1), self._OPERATOR_REPR[operator],
                to_str(operand2), str(answer))

    def _next_question(self):
        self._question = _Question(self._generate_question())
        self._keypad.text = ''.join(self._question.value[:3])
        self._num_questions_asked += 1

    @property
    def completed(self):
        return (self._question and
                self._question.value == self._SPECIAL_QUESTION and
                self._keypad.text == _QuestionState.OK.value)

    def start(self):
        assert not self._active
        self._active = True
        self._keypad.text_color = color.RED
        pygame.time.set_timer(self._TICK, self._FREQ_MS)

    def stop(self):
        assert self._active
        self._init_state()
        self._keypad.set_initial_text()
        self._door.text = self._keypad.text
        pygame.time.set_timer(self._TICK, 0)

    def send(self, event):
        if not self._active:
            self.start()
            return True
        elif event.type == self._TICK:
            if self._keypad.text == _QuestionState.ERR.value:
                self.stop()
                return True
            elif self.completed:
                return True
            elif (not self._question or
                  self._keypad.text == _QuestionState.OK.value):
                self._next_question()
            elif self._question.state is not _QuestionState.ACTIVE:
                self._keypad.text = self._question.state.value
            else:
                self._question.tick()
            self._flip_text_color()
            return True
        elif event.type == KEYDOWN and event.unicode.isdigit():
            if self._question:
                self._question.solve(event.unicode)
            return True
        return False

    def _flip_text_color(self):
        if self._keypad.text_color == color.BLACK:
            self._keypad.text_color = color.RED
        else:
            self._keypad.text_color = color.BLACK
